#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "toolkit.h"
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"
#include "CdiJni.h"

int cdiSwapBytes(char *dat,int size,int fmt);

JavaVM *gJVM = NULL;
JNIEnv *gJniEnv = NULL;
typedef struct CdiCbElemStruct
{
  int cdiId; // as returned from cdiAttachLink (needed for closing the link)
  jint cbId; // requested callback id in java
  jobject cbTLink; // reference to callback object in java
  jmethodID cbMid; // callback method id
  jbyteArray dataRef; // reference to the data object in java
  DTYPE dtype; // reference to the data object in cdi
} CdiCbElem;
#define MAX_CDI_LINK_ELEMS 1024
int gMaxCdiLinkElems = MAX_CDI_LINK_ELEMS;
CdiCbElem *cdiCbElems = NULL;
#define MAX_INPUT_SIZE 4096
int inputBufferSize = MAX_INPUT_SIZE;
BYTE inputBuffer[MAX_INPUT_SIZE];

int initLinkElemTable(void)
{
  int i;
  if ((cdiCbElems=(CdiCbElem *)calloc(gMaxCdiLinkElems,sizeof(CdiCbElem))) == NULL)
  {
    cdilog("JNI: cannot allocate memory for callback elements table!");
    return out_of_server_memory;
  }
  for (i=0; i<gMaxCdiLinkElems; i++) cdiCbElems[i].cdiId = -1;
  return 0;
}
int getNextElemId(void)
{
  int i;
  if (cdiCbElems == NULL) return -out_of_server_memory;
  for (i=0; i<gMaxCdiLinkElems; i++) if (cdiCbElems[i].cdiId == -1) break;
  if (i == gMaxCdiLinkElems) return -resources_exhausted;
  return i;
}
int addLinkElem(int cdiId,jobject cb,jmethodID cbMid,jint cbId,jbyteArray dataRef,DTYPE *dout)
{
  int i;
  if (cdiId < 0) return cdiId;
  if ((i=getNextElemId()) >= 0)
  {
    cdiCbElems[i].cdiId = cdiId;
    cdiCbElems[i].cbTLink = cb;
    cdiCbElems[i].cbMid = cbMid;
    cdiCbElems[i].cbId = cbId;
    cdiCbElems[i].dataRef = dataRef;
    cdiCbElems[i].dtype = *dout;
  }
  return i;
}
int rmvLinkElem(jint linkId)
{
  if (cdiCbElems == NULL) return out_of_server_memory;
  if (linkId < 0 || linkId >= gMaxCdiLinkElems) return link_not_open;
  memset(&cdiCbElems[linkId],0,sizeof(CdiCbElem));
  cdiCbElems[linkId].cdiId = -1;
  return 0;
}

static int gCdiDebugLevel = 0;
void dumpData(char *txt, DTYPE *dt)
{
  int i;
  char tmp[32],str[1024];
  sprintf(str,"%s: ",txt);
  puts(str);
  str[0] = 0;
  switch (dt->dFormat)
  {
    default:
    case CF_BYTE:
      for (i=0; i<(int)dt->dArrayLength; i++)
      {
        sprintf(tmp,"0x%x ",dt->data.bptr[i]);
        if (strlen(str) < 1016) strcat(str,tmp); else break;
      }
      break;
    case CF_SHORT:
      for (i=0; i<(int)dt->dArrayLength; i++)
      {
        sprintf(tmp,"%d ",dt->data.sptr[i]);
        if (strlen(str) < 1016) strcat(str,tmp); else break;
      }
      break;
    case CF_LONG:
      for (i=0; i<(int)dt->dArrayLength; i++)
      {
        sprintf(tmp,"%d ",dt->data.lptr[i]);
        if (strlen(str) < 1008) strcat(str,tmp); else break;
      }
      break;
    case CF_TEXT:
      strncpy(str,dt->data.cptr,1024);
      break;
  }
  puts(str);
  return;
}
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm,void *reserved) 
{
  JNIEnv *env = NULL;
  gJVM = vm;
  (*gJVM)->GetEnv(gJVM,(void **)&env,JNI_VERSION_1_2);
  return JNI_VERSION_1_2;
}
JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *vm,void *reserved) 
{
  cdilog("JNI: JNI_OnUnload called");
}

void qdCallback(int id,int cc) 
{
  int siz;
  jmethodID mid;
  jobject jCb;
  static JNIEnv *env = NULL;
  static int isAttached = FALSE;
  static int callCounter = 0;

  // this callback will be called from a different thread than the 'AttachLink' thread !
  if (gJVM == NULL) return;
  if (!isAttached)
  {
    (*gJVM)->AttachCurrentThread(gJVM,(void **)&env,NULL);  
     isAttached = TRUE;
  }
  if (env == NULL) return;
  if (id < 0 || id > gMaxCdiLinkElems) 
  {
    cdilog("JNI: invalid callback id : %d",id);
    return; // this can't be !
  }
  if ((jCb=cdiCbElems[id].cbTLink) != NULL && cdiCbElems[id].dtype.data.bptr != NULL)
  {
    mid = cdiCbElems[id].cbMid;
    if (gCdiDebugLevel > 0) dumpData("read from bus",&cdiCbElems[id].dtype);
#   ifdef INTEL
    /* swap the little-endian intel bytes to java order ... */
    cdiSwapBytes(cdiCbElems[id].dtype.data.bptr,cdiCbElems[id].dtype.dArrayLength,cdiCbElems[id].dtype.dFormat);
#   endif
    siz = cdiCbElems[id].dtype.dArrayLength * fmtsizeof(cdiCbElems[id].dtype.dFormat);
    (*env)->SetByteArrayRegion(env,cdiCbElems[id].dataRef,0,siz,cdiCbElems[id].dtype.data.bptr);
    (*env)->CallVoidMethod(env,jCb,mid,cdiCbElems[id].cbId,cc);
    callCounter++;
  }
  else
  {
    cdilog("JNI: callback error! java cb addr : %x, java data addr : %x",jCb,cdiCbElems[id].dtype.data.bptr);
  }
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_closeLink(JNIEnv *env,jobject obj,jint id)
{
  jint cc = 0;
  jbyteArray dataRef;
  BYTE *bDataRef;

  if (cdiCbElems[id].cdiId < 0) return link_not_open;
  cc = cdiCloseLink(cdiCbElems[id].cdiId);
  dataRef = cdiCbElems[id].dataRef;
  bDataRef = cdiCbElems[id].dtype.data.bptr;
  if (dataRef != NULL) (*env)->ReleaseByteArrayElements(env,dataRef,bDataRef,0);
  (*env)->DeleteGlobalRef(env,dataRef);
  (*env)->DeleteGlobalRef(env,cdiCbElems[id].cbTLink);
  rmvLinkElem(id);
  cdilog("JNI: close cdi link %d : %s",id,erlst[cc]);
  return cc;
}
JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_attachLink(
  JNIEnv *env, 
  jobject obj, 
  jstring jdevName, 
  jstring jdevProperty, 
  jshort outDataType, 
  jint outDataLength, 
  jbyteArray outDataRef, 
  jshort inDataType, 
  jint inDataLength, 
  jbyteArray inDataRef, 
  jshort accessFlags, 
  jint pollingInterval, 
  jobject cbCdi, 
  jint accessMode, 
  jint cbId
  )
{
  DTYPE dout, din;
  BYTE *bytes, *inputBytes;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  const char *devProperty = (*env)->GetStringUTFChars(env,jdevProperty,0);
  int siz, idCdi, id = -not_implemented;
  jmethodID cbMid;
  jclass cls;
  jobject gCbCdi,gDataRef;

  if (gCdiDebugLevel > 0) printf("JNI: cdiAttachLink %s %s\n",devName,devProperty);
  dout.dFormat = LFMT(outDataType);
  dout.dArrayLength = (UINT32)outDataLength;
  dout.dTag[0] = 0;
  dout.data.bptr = NULL;
  if (outDataRef != NULL) dout.data.bptr = (*env)->GetByteArrayElements(env,outDataRef,0); // ?? want to write into this reference !

  din.dFormat = LFMT(inDataType);
  din.dArrayLength = (UINT32)inDataLength;
  din.dTag[0] = 0;
  din.data.bptr = NULL;
  if (inDataRef != NULL) 
  {
    inputBytes = bytes = (*env)->GetByteArrayElements(env,inDataRef,0);
#   ifdef INTEL
    /* swap the big-endian java bytes to intel order ... */
    siz = din.dArrayLength*fmtsizeof(din.dFormat);
    if (siz < MAX_INPUT_SIZE) 
    {
      inputBytes = inputBuffer; // this should almost alway work
      memcpy(inputBuffer,bytes,siz);
    }
    cdiSwapBytes(inputBytes,din.dArrayLength,din.dFormat);
#   endif
    din.data.bptr = inputBytes;
    if (gCdiDebugLevel > 0) dumpData("write to bus", &din);
  }

  gJniEnv = env;
  cls = (*env)->GetObjectClass(env,cbCdi);
  cbMid = (*env)->GetMethodID(env,cls,"callback", "(II)V");
  id = getNextElemId();
  idCdi = cdiAttachLinkEx((char *)devName,(char *)devProperty,&dout,&din,accessFlags,pollingInterval,qdCallback,accessMode,id);
  gCbCdi = (*env)->NewGlobalRef(env,cbCdi);
  gDataRef = (*env)->NewGlobalRef(env,outDataRef);
  id = addLinkElem(idCdi,gCbCdi,cbMid,cbId,gDataRef,&dout);

  if (inDataRef != NULL) (*env)->ReleaseByteArrayElements(env,inDataRef,bytes,0);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  (*env)->ReleaseStringUTFChars(env,jdevProperty,devProperty);

  return id;
}
JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_execLink(
  JNIEnv *env, 
  jobject obj, 
  jstring jdevName, 
  jstring jdevProperty, 
  jshort outDataType, 
  jint outDataLength, 
  jbyteArray outDataRef, 
  jshort inDataType, 
  jint inDataLength, 
  jbyteArray inDataRef, 
  jshort accessFlags, 
  jint timeoutInterval 
  )
{
  DTYPE dout, din;
  BYTE *bytes, *inputBytes;
  int siz;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  const char *devProperty = (*env)->GetStringUTFChars(env,jdevProperty,0);
  jint cc = not_implemented;

  dout.dFormat = LFMT(outDataType);
  dout.dArrayLength = (UINT32)outDataLength;
  dout.dTag[0] = 0;
  dout.data.bptr = NULL;
  if (outDataRef != NULL) dout.data.bptr = (*env)->GetByteArrayElements(env,outDataRef,0); 

  din.dFormat = LFMT(inDataType);
  din.dArrayLength = (UINT32)inDataLength;
  din.dTag[0] = 0;
  din.data.bptr = NULL;
  if (inDataRef != NULL) 
  {
    inputBytes = bytes = (*env)->GetByteArrayElements(env,inDataRef,0);
#   ifdef INTEL
    /* swap the big-endian java bytes to intel order ... */
    siz = din.dArrayLength*fmtsizeof(din.dFormat);
    if (siz < MAX_INPUT_SIZE) 
    {
      inputBytes = inputBuffer; // this should almost always work
      memcpy(inputBuffer,bytes,siz);
    }
    cdiSwapBytes(inputBytes,din.dArrayLength,din.dFormat);
#   endif
    din.data.bptr = inputBytes;
    if (gCdiDebugLevel > 0) dumpData("write to bus", &din);
  }

  cc = cdiExecLinkEx((char *)devName,(char *)devProperty,&dout,&din,accessFlags,(UINT16)timeoutInterval);

  if (inDataRef != NULL) (*env)->ReleaseByteArrayElements(env,inDataRef,bytes,0);
  if (outDataRef != NULL)
  {
    if (gCdiDebugLevel > 0) dumpData("read from bus",&dout);
#   ifdef INTEL
    /* swap the little-endian intel bytes to java order ... */
    cdiSwapBytes(dout.data.bptr,dout.dArrayLength,dout.dFormat);
#   endif
    (*env)->ReleaseByteArrayElements(env,outDataRef,dout.data.bptr,JNI_COMMIT);
  }
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  (*env)->ReleaseStringUTFChars(env,jdevProperty,devProperty);

  return cc;
}
JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getDeviceNames(JNIEnv *env,jobject obj,jbyteArray namebuffer,jint bufsize)
{
  jint cc = 0;
  jbyte *jnamebuffer;
  
  if (namebuffer == NULL || bufsize < 1) return cdi_getDeviceNames((NAME16 *)NULL,0);
  jnamebuffer = (*env)->GetByteArrayElements(env,namebuffer,0); 
  cc = cdi_getDeviceNames((NAME16 *)jnamebuffer,bufsize);
  (*env)->ReleaseByteArrayElements(env,namebuffer,jnamebuffer,JNI_COMMIT);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getBusNames(JNIEnv *env,jobject obj,jbyteArray namebuffer,jint bufsize)
{
  jint cc = 0;
  jbyte *jnamebuffer;
  
  if (namebuffer == NULL || bufsize < 1) return cdi_getBusNames((NAME16 *)NULL,0);
  jnamebuffer = (*env)->GetByteArrayElements(env,namebuffer,0); 
  cc = cdi_getBusNames((NAME16 *)jnamebuffer,bufsize);
  (*env)->ReleaseByteArrayElements(env,namebuffer,jnamebuffer,JNI_COMMIT);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getBusName(JNIEnv *env,jobject obj,jstring jdevName,jbyteArray namebuffer)
{
  jint cc = 0;
  jbyte *jnamebuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  char *busname;
  
  if (namebuffer == NULL || jdevName == NULL) return argument_list_error;
  jnamebuffer = (*env)->GetByteArrayElements(env,namebuffer,0); 

  busname = cdi_getBusNameForDevice((char *)devName);
  if (busname != NULL) memcpy(jnamebuffer,busname,16);

  (*env)->ReleaseByteArrayElements(env,namebuffer,jnamebuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getBusAddr(JNIEnv *env,jobject obj,jstring jdevName,jintArray addrbuffer)
{
  jint cc = 0;
  jint *jaddrbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  DADDRESS *addr;
  
  if (addrbuffer == NULL || jdevName == NULL) return argument_list_error;
  jaddrbuffer = (*env)->GetIntArrayElements(env,addrbuffer,0); 

  addr = cdi_getBusAddressForDevice((char *)devName);
  /* GetIntArrayElements() seems to do the swapping to java order ... */
  if (addr != NULL) 
  {
    memcpy(jaddrbuffer,addr,4*sizeof(int));
  }
  (*env)->ReleaseIntArrayElements(env,addrbuffer,jaddrbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getCdiNumber(JNIEnv *env,jobject obj,jstring jdevName,jintArray addrbuffer)
{
  jint cc = 0;
  jint *jaddrbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  int cdiDevNr, cdiLine, cdiIndex;
  
  if (addrbuffer == NULL || jdevName == NULL) return argument_list_error;
  jaddrbuffer = (*env)->GetIntArrayElements(env,addrbuffer,0); 

  cdiDevNr = findDevice((char *)devName,&cdiLine,&cdiIndex);
  jaddrbuffer[0] = cdiDevNr;
  jaddrbuffer[1] = cdiLine;
  jaddrbuffer[2] = cdiIndex;
  (*env)->ReleaseIntArrayElements(env,addrbuffer,jaddrbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getRange(JNIEnv *env,jobject obj,jstring jdevName,jfloatArray fltbuffer)
{
  jint cc = 0;
  jfloat *jfltbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  float flt[2];
  
  if (fltbuffer == NULL || jdevName == NULL) return argument_list_error;
  jfltbuffer = (*env)->GetFloatArrayElements(env,fltbuffer,0); 
  
  cdi_getRangeForDevice((char *)devName,flt);
  jfltbuffer[0] = flt[0];
  jfltbuffer[1] = flt[1];

  (*env)->ReleaseFloatArrayElements(env,fltbuffer,jfltbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_scanbus(JNIEnv *env,jobject obj,jstring jbusname, jstring jtarget,jbyteArray jresult,jint jbufsize)
{
  jint cc = 0;
  const char *busname = (*env)->GetStringUTFChars(env,jbusname,0);
  const char *target = (*env)->GetStringUTFChars(env,jtarget,0);
  jbyte *resultBuffer;
  
  if (busname == NULL || target == NULL || jresult == NULL || jbufsize < 1) 
    return argument_list_error;

  resultBuffer = (*env)->GetByteArrayElements(env,jresult,0); 

  cc = cdiScanBus((char *)busname,(char *)target,(char *)resultBuffer,jbufsize);

  (*env)->ReleaseByteArrayElements(env,jresult,resultBuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jbusname,busname);
  (*env)->ReleaseStringUTFChars(env,jtarget,target);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getTemplateNames(JNIEnv *env,jobject obj,jbyteArray namebuffer,jint bufsize)
{
  jint cc = 0;
  jbyte *jnamebuffer;
  
  if (namebuffer == NULL || bufsize < 1) return cdi_getBusNames((NAME16 *)NULL,0);
  jnamebuffer = (*env)->GetByteArrayElements(env,namebuffer,0); 
  cc = cdi_getTemplateNames((NAME16 *)jnamebuffer,bufsize);
  (*env)->ReleaseByteArrayElements(env,namebuffer,jnamebuffer,JNI_COMMIT);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getTemplate(JNIEnv *env,jobject obj,jstring jtemplateName,jbyteArray jtemplateBuffer,jint bufsize)
{
  jint cc = 0, i, siz;
  const char *templName = (*env)->GetStringUTFChars(env,jtemplateName,0);
  jbyte *templateBuffer;
  NAME16II *templ;
  
  if (jtemplateBuffer == NULL || jtemplateName == NULL || bufsize == 0) return argument_list_error;
  templateBuffer = (*env)->GetByteArrayElements(env,jtemplateBuffer,0); 

  siz = bufsize; templ = (NAME16II *)templateBuffer;
  cc = cdi_getTemplate((char *)templName,templ,&siz);

  if (cc == 0) 
  {
    if (siz > bufsize) return buffer_too_small;
#   ifdef INTEL
    /* swap the little-endian intel bytes to java order ... */
    for (i=0; i<siz; i++)
    {
      cdiSwapBytes((BYTE *)&templ[i].i1val,2,CF_LONG);
    }
#   endif
  }
  (*env)->ReleaseByteArrayElements(env,jtemplateBuffer,templateBuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jtemplateName,templName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getDescription(JNIEnv *env,jobject obj,jstring jdevName,jbyteArray descbuffer)
{
  jint cc = 0;
  jbyte *jdescbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  char desc[64];
  
  if (descbuffer == NULL || jdevName == NULL) return argument_list_error;
  jdescbuffer = (*env)->GetByteArrayElements(env,descbuffer,0); 

  desc[0] = 0;
  cc = cdi_getDescriptionForDevice((char *)devName,desc);
  strncpy(jdescbuffer,desc,64);

  (*env)->ReleaseByteArrayElements(env,descbuffer,jdescbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getPrecision(JNIEnv *env,jobject obj,jstring jdevName,jbyteArray strbuffer)
{
  jint cc = 0;
  jbyte *jstrbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  char str[64];
  
  if (strbuffer == NULL || jdevName == NULL) return argument_list_error;
  jstrbuffer = (*env)->GetByteArrayElements(env,strbuffer,0); 

  str[0] = 0;
  cc = cdi_getPrecisionForDevice((char *)devName,str);
  strncpy(jstrbuffer,str,64);

  (*env)->ReleaseByteArrayElements(env,strbuffer,jstrbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getUnits(JNIEnv *env,jobject obj,jstring jdevName,jbyteArray strbuffer)
{
  jint cc = 0;
  jbyte *jstrbuffer;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  char str[64];
  
  if (strbuffer == NULL || jdevName == NULL) return argument_list_error;
  jstrbuffer = (*env)->GetByteArrayElements(env,strbuffer,0); 

  str[0] = 0;
  cc = cdi_getUnitsForDevice((char *)devName,str);
  strncpy(jstrbuffer,str,64);

  (*env)->ReleaseByteArrayElements(env,strbuffer,jstrbuffer,JNI_COMMIT);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return cc;
}

JNIEXPORT jint JNICALL Java_de_desy_cdi_CdiJni_getBusErrorCount(JNIEnv *env,jobject obj,jstring jdevName)
{
  jint cc = 0;
  const char *devName = (*env)->GetStringUTFChars(env,jdevName,0);
  int errcnt = 0;
  
  if (jdevName == NULL) return argument_list_error;

  errcnt = cdi_getDeviceNumberOfErrors((char *)devName);
  (*env)->ReleaseStringUTFChars(env,jdevName,devName);
  return errcnt;
}

BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  int cc;
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      cc = initLinkElemTable();
      cdilog("JNI: cdi java native interface loaded : %s",erlst[cc]);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      //cdiCloseLinks();
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      cdiCloseLinks();
      break;
  }
  return (TRUE);
}
