#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include "tinecdi.h"
#include "cdi.h"
#include "cdiSedacPKTRLib.h"

#define WAIT_TIME_NS 1000000  // 1ms
int nSedacLines = N_SEDAC_LINES;
int SedacSimulate = 0;

SEDAC_STRUCT sedacInfo[N_SEDAC_LINES] = {
	{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1},  {-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}
};

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodules","writemodules"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodules, writemodules, nCommands};

__declspec(dllexport) int sedacScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
  CdiLineInfoBlk *busHandle;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  
  if (j == getcommands)
  {
    _snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }
  
  if ( j == getlines )
  {
    _snprintf(result,resultBufferSize,"Sedac lines configured:\n");
    for (i=0, cnt=0; i<N_SEDAC_LINES; i++)
    {
      if ( sedacInfo[i].sedac_fd < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       _snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > nSedacLines || sedacInfo[line-1].sedac_fd < 0) return no_such_line;

  if ((busHandle=cdiGetBusHandle(sedacInfo[line-1].cdiLine)) == NULL)
  {
    cdilog("sedpc: Wrong cdi_line: %d for busscan() call\n",sedacInfo[line-1].cdiLine);
    return no_such_line;;
  }

  if ( j == getcrates )
  {
    _snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
    cdiMutexLock(&busHandle->lineMutex);
    for (crate=0, cnt=0; crate<32; crate++)
    {
      ret = sedac(line, crate, 253, CA_READ, &rData, 1, NULL);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
      cnt++;
    }
    cdiMutexUnlock(&busHandle->lineMutex);
    return 0;
  }
  else if ( j == getmodules )
  {
    crate = data[1];
    if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
    _snprintf(result,resultBufferSize,"Subaddress on line %d crate %d was found:\n",line,crate);
    cdiMutexLock(&busHandle->lineMutex);
    for (subaddr=0, cnt=0; subaddr<256; subaddr++)
    {
      ret = sedac(line, crate, subaddr, CA_READ, &rData, 1, NULL);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
      cnt++;
    }
    cdiMutexUnlock(&busHandle->lineMutex);
    return 0;
  }
  crate = data[1];
  subaddr = data[2];
  if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
  if ( j == readmodules )
  {
    _snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle->lineMutex);
    ret = sedac(line, crate, subaddr, CA_READ, &rData, 1, NULL);
    cdiMutexUnlock(&busHandle->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
  }
  else if ( j == writemodules )
  {
    wData = data[3];
    if ( nData < 4 ) return argument_list_error;
    _snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle->lineMutex);
    ret = sedac(line, crate, subaddr, CA_WRITE, &wData, 1, NULL);
    cdiMutexUnlock(&busHandle->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
  }  
  return 0;
}

__declspec(dllexport) int filterSedacBus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return -argument_list_error;
  return 0;
}
__declspec(dllexport) int initSedacForCDI(int busLine, int cdiLine, int numberDevices,char *parameterList)
{ int ret;

  ret = initSedac(busLine);
  if ( ret >= 0 ) sedacInfo[busLine - 1].cdiLine = cdiLine;
  return ret;
}
__declspec(dllexport) int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ret, length, writeOffset;
  void *pOut;

// pktr sedac driver is a kernel driver, no need of mutex for protection!!!
//  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) ? CA_WRITE : CA_READ;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? CA_WRITE : CA_READ;
  else
    access = CA_WRITE;

  if ( access == CA_WRITE )
  {
    if ( pReq->accessFlag == cdiWriteWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag )
    {
      if ( step == 1 )
      {
        length = sizeof(short);
        pOut = (void *)pReq->pOutData;
        writeOffset = sedacSubOffsetSecond;
      }
      else
      {
        length = (pReq->accessFlag == cdiWriteReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : (pReq->outLength - sizeof(short));
        pOut = (void *)((char *)pReq->pOutData + sizeof(short));
        writeOffset = sedacSubOffsetThird;
      }
    }
    else
    {
      length = (pReq->accessFlag == cdiReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : pReq->outLength;
      pOut = (void *)pReq->pOutData;
      writeOffset = sedacSubOffsetSecond;
    }
  }

  ret = sedac((WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (WORD)pReq->pDev->cdiAddr,
//          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[access == CA_READ ? sedacSubOffset : writeOffset]), 
          (int)access,
          (WORD *)(access == CA_READ ? pReq->pInData : pOut),
          (int)(access == CA_READ ? pReq->inLength : length) / sizeof(short),
          pReq );
  if ( ret != 0 )
  {
    pReq->errorCode = ret < 0 ? -ret : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }
//  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  return pReq->errorCode;
}

/********************************************************/
// if sync call, data not pass filter then repeat Transmit
__declspec(dllexport) int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}
/********************************************************/
/* only call once with pReq->active == transRequest */
/********************************************************/
__declspec(dllexport) void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  do
  {
    doSedacRequest(pReq, 1);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest ) return;
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  do
  {
    doSedacRequest(pReq, 2);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag)
  {
    pReq->active = finishRequest; /* finish request */
    return;
  }

  /* do third part */
  doSedacRequest(pReq, 3);
  pReq->active = finishRequest; /* finish request */
}

/* ============  init routine  =================================== */
__declspec(dllexport) int initSedac(int ln)
{ char fname[80];
  int i;
  SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return ln;		// ln number (0, 1, ...) as file handle!

  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd >= 0 )
  {
    cdilog("sedac already opened! sedac line: %d!\n", ln+1);
    return pSedac->sedac_fd;
  }
	
  sprintf(fname,"\\\\.\\ntSedac%d",ln+1);
  pSedac->sedacHandle = CreateFile(
			  fname,			// Open the Device "file"
			  GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_OVERLAPPED,  //      0,
        NULL
  );
  pSedac->sedac_fd = (int)pSedac->sedacHandle;
	if ( pSedac->sedacHandle == INVALID_HANDLE_VALUE)
  {
		cdilog( "initSedac for PKTR card CreateFile error! line: %d; GetLastError: 0x%x", ln+1, GetLastError() );
    pSedac->sedac_fd = -1;
		return -file_not_open;
  }

  for (i=0; i<N_SED_BUFFERS; i++)
  {
    pSedac->locBuffer[i].da = NULL;
    pSedac->locBuffer[i].dataSize = 0;	
    pSedac->locBuffer[i].bufFlag = 0;	
  }
  pSedac->sedacSimulationCount = 0;

  cdilog("Open pktr sedac line OK! line: %d\n",ln+1);	
  return pSedac->sedac_fd;
}

// ============  exit routine  ===================================
__declspec(dllexport) int exitSedac(int ln)
{ SEDAC_STRUCT *pSedac;
  int i;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return 0;

  cdilog("close pktr sedac line! line number: %d\n", ln+1);
  
  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd < 0 ) return -file_not_open;
  
  close(pSedac->sedac_fd);
  pSedac->sedac_fd = -1;

  for (i=0; i<N_SED_BUFFERS; i++) if ( pSedac->locBuffer[i].da )
  {
    free(pSedac->locBuffer[i].da);
    pSedac->locBuffer[i].da = NULL;
    pSedac->locBuffer[i].dataSize = 0;	
    pSedac->locBuffer[i].bufFlag = 0;	
  }
  pSedac->sedacSimulationCount = 0;

  return 0;
}

// ============  getBuffer  ===================================
PSEDAC_ARY getBuffer(int ln, int number, int index)
{ SEDAC_STRUCT *pSedac;

  pSedac = &sedacInfo[ln];
  if ( pSedac->locBuffer[index].da )
  {
    free(pSedac->locBuffer[index].da);
    pSedac->locBuffer[index].da = NULL;
  }

  pSedac->locBuffer[index].da = (PSEDAC_ARY)malloc( sizeof(SEDAC_ARY) * (number + 1) );
  if ( pSedac->locBuffer[index].da != NULL )
  {
    pSedac->locBuffer[index].dataSize = number;
    pSedac->locBuffer[index].bufFlag = 0;
  }
  else
  {
    cdilog("sedac pktr line: %d; buffer index: %d;  => get buffer failed!! size: %d\n",ln+1,index,number);
    return NULL;
  }
  cdilog("sedac pktr line: %d; buffer index: %d;  => get buffer: %x; size: %d\n",ln+1,index,(int)(pSedac->locBuffer[index].da),number);
  return pSedac->locBuffer[index].da;
}

int execSedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords, int flag, CdiRequestInfoBlk *pReq)
{ int i, ret, NumAddr, err, index;
  PSEDAC_DATA_HEADER phe;
  SEDAC_STRUCT *pSedac;
  PSEDAC_ARY buf;
  ULONG DataLength,InputDataLength;
  DWORD RetLength;
  
  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  pSedac = &sedacInfo[ln];

  if ( SedacSimulate )
  {
    i = nwords * 250;
    if ( i > 4 ) millisleep ( i/4 ); else millisleep(1);
    for (i=0; i<nwords; i++, data++)
    {
      ret = flag ? RWF[i] : RWF[0];
      if ( ret == CA_READ )
      *data = pSedac->sedacSimulationBuffer + pSedac->sedacSimulationCount++;
      else
      pSedac->sedacSimulationBuffer = *data;
    }
    return 0;
  }

  if ( pSedac->sedac_fd < 0 ) return -file_not_open;

  // find free buffer
  for (index=0; index<N_SED_BUFFERS; index++) if ( pSedac->locBuffer[index].bufFlag == 0 ) break;
  if ( index < N_SED_BUFFERS )
  {
    pSedac->locBuffer[index].bufFlag = 1;
    if ( pReq ) pReq->bufIndex = index;
  }
  else
  {
    cdilog("No free buffer slot for Sedac!! process pointer: 0x%x; sedac line: %d\n",pReq, ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine);
    return -resources_exhausted;
  }

  if ( (pSedac->locBuffer[index].da == NULL || pSedac->locBuffer[index].dataSize < nwords)
  	&& getBuffer(ln, nwords, index) == NULL )
  {
    if ( pReq ) pReq->bufIndex = -1;
    pSedac->locBuffer[index].bufFlag = 0;
    return -out_of_local_memory;
  }

  phe = (PSEDAC_DATA_HEADER)pSedac->locBuffer[index].da;
  phe->SedacLine = ln;	// fill header (2 short only)
  phe->NumberOfError = 0;
  buf = pSedac->locBuffer[index].da;
  buf++;

//  cdilog("execSedac: flag = %d; line: %d; crate: %d; sub: %d; RWF: %d; nwordds: %d\n",flag,ln+1,crate[0],subadd[0], RWF[0], nwords);
  for (i=0; i<nwords; i++) 
  {
    ret = flag ? RWF[i] : RWF[0];
  	if ( ret == CA_WRITE ) buf[i].Data = data[i];
    buf[i].CommandReturn = (ret == CA_READ ? SEDAC_READ : SEDAC_WRITE) | (((flag ? crate[i] : crate[0]) << 8) & CRATE_ADR) | ((flag ? subadd[i] : subadd[0]) & SUB_ADR);
  }
  NumAddr = nwords;
//  NumAddr = (flag == 0 && RWF[0] == CA_READ) ? 1 : nwords; // same address and only read, input one telegram for read same address! need load address once only

  DataLength = (nwords + 1) * sizeof(ULONG); // plus header
  InputDataLength = (NumAddr + 1) * sizeof(ULONG);
	RetLength = 0;

  // Make IOCTL call. It should return TRUE, if successfull
  ret = DeviceIoControl(
					pSedac->sedacHandle,   	// Handle to device
					IOCTL_SEDAC_DATA,     // IO Control code for Write
					(PULONG)phe,   // Buffer to driver.  Holds port & data.
					InputDataLength,      // Length of buffer in bytes.
					(PULONG)phe,   // Buffer from driver.   Not used.
					DataLength,           // Length of buffer in bytes.
					&RetLength,        		// Bytes placed in outbuf.  Should be 0.
					NULL						  		// NULL means wait till I/O completes.
				);
//  cdilog("execSedac: flag = %d; line: %d; crate: %d; sub: %d; RWF: %d; nwordds: %d; ret: %d; addr: 0x%x\n",flag,ln+1,crate[0],subadd[0], RWF[0], nwords, ret, data);

  // Did the IOCTL succeed?
  if (ret && RetLength == DataLength )
    for (err=0, i=0; i<nwords; i++)    //fill errors => return
  	{
      if ( (flag ? RWF[i] : RWF[0]) == CA_READ ) data[i] = buf[i].Data;
	    if ( buf[i].CommandReturn ) err++;
	  }
  else
  {
    cdilog("pktr sedac DeviceIoControl error: line: %d, getLastError: 0x%x\n",ln+1, GetLastError());
    err = -software_error;
  }
  if ( pReq ) pReq->bufIndex = -1;
  pSedac->locBuffer[index].bufFlag = 0;
  return err;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(u16 ln, u16 crate, u16 subadd, int RWF, u16 *data, int nwords, CdiRequestInfoBlk *pReq)
{
  return execSedac(ln, &crate, &subadd, &RWF, data, nwords, 0, pReq);
}

// ================  read/write nwords telegram to different addresses ===================================
//int nsedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords)
//{
//  return execSedac(ln, crate, subadd, RWF, data, nwords, 1);
//}

