#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

//#include <iostream.h>
//#include <windows.h>
//#include <conio.h>
//#include "c:\\twincat\\ads api\\tcadsdll\\include\\tcadsdef.h"
//#include "c:\\twincat\\ads api\\tcadsdll\\include\\tcadsapi.h"
#include "tcadsdef.h"
#include "tcadsapi.h"

#include "tinecdi.h"
#include "cdi.h"
#include "twinCATadsLib.h"

#define N_TWINCAT_LINES 8
#define MAX_SEDAC_PUT 350   /* number of telegrams to put (4k FIFO 680 telegrams) */
int nTwinCATLines = N_TWINCAT_LINES;
#define WAIT_TIME_NS 10000000  // 10ms
typedef struct
{
  int initFlag, cdiLine;
  long      nErr, nPort;  // local port number
  AmsAddr   Addr;         // remote SPS port & ip
  PAmsAddr  pAddr;
} TWINCAT_DATA;
TWINCAT_DATA twincatInfo[N_TWINCAT_LINES] = { /* init */
  {0}, {0}, {0}, {0},  {0}, {0}, {0}, {0}
};
char *scanCommand[] = {"getcommands","adsstatus","startplc","stopplc"};
enum scommand {getcommands, adsstatus, startplc, stopplc, nCommands};

int reqHandleTime=0, rwAdsTime=0, totalAdsTime=0;

__declspec(dllexport) int twinCatScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, nPort, nErr, spsPort, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
  CdiLineInfoBlk *busHandle;
  AmsAddr   Addr;
  SHORT  nAdsState;
  USHORT  nDeviceState;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  if (j == getcommands)
  {
    _snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }

  if ( j == adsstatus || j == startplc || j == stopplc)
  {
    if ( nData > 7 || (nData < 4 && nData > 1) ) return argument_list_error;
    if ( j == adsstatus )
      _snprintf(result,resultBufferSize,"ADS Status:\n");
    else if ( j == startplc )
      _snprintf(result,resultBufferSize,"Start PLC:\n");
    else
      _snprintf(result,resultBufferSize,"Stop PLC:\n");
    nPort = AdsPortOpenEx();
    if ( nPort == 0 )
    {
      slen = strlen(result);
      strncat(result,"AdsPortOpenEx() failed!",resultBufferSize-slen);
      return 0;
    }
    if ( nData > 1 ) // open local address
    {
      cnt = nData < 6 ? 4 : 6;
      for (i=0; i<6; i++) Addr.netId.b[i] = (cnt == 6 || i < 4) ? data[i] : 1;
      spsPort = (nData % 2) ? data[cnt] : 0;
    }
    else
    {
      nErr = AdsGetLocalAddressEx(nPort, &Addr);
      if ( nErr )
      {
        if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"AdsGetLocalAddressEx failed: %d",nErr);
        AdsPortCloseEx(nPort);
        return 0;
      }
      spsPort = nData ? data[0] : 0;
    }
    cnt = spsPort <= 0 ? 4 : 1;   // check all 4 sps ports
    for ( i=0; i<cnt; i++)
    {
      if ( cnt == 1 ) Addr.port = spsPort;
      else
      {
        if ( i == 0 ) Addr.port = AMSPORT_R0_PLC_RTS1;
        else if ( i == 1 ) Addr.port = AMSPORT_R0_PLC_RTS2;
        else if ( i == 2 ) Addr.port = AMSPORT_R0_PLC_RTS3;
        else Addr.port = AMSPORT_R0_PLC_RTS4;
      }
      if ( j == adsstatus )
      {
        nErr = AdsSyncReadStateReqEx(nPort,(PAmsAddr)&Addr, &nAdsState, &nDeviceState);
        if ((slen = strlen(result)) < resultBufferSize - 18)
          if ( nErr )
            _snprintf(&result[slen],resultBufferSize-slen,"AdsSyncReadStateReq failed: %d; port: %d\n",nErr, Addr.port);
          else
            _snprintf(&result[slen],resultBufferSize-slen,"Port: %d; AdsState: %d; DeviceState: %d\n",Addr.port, nAdsState, nDeviceState);
      }
      else
      {
        nAdsState = j == startplc ? ADSSTATE_RUN : ADSSTATE_STOP;
        nDeviceState = 0;
        nErr = AdsSyncWriteControlReqEx(nPort,(PAmsAddr)&Addr, nAdsState, nDeviceState, 0, NULL);
        if ((slen = strlen(result)) < resultBufferSize - 18)
          if ( nErr )
            _snprintf(&result[slen],resultBufferSize-slen,"AdsSyncWriteControlReqEx failed: %d; port: %d\n",nErr, Addr.port);
          else
            _snprintf(&result[slen],resultBufferSize-slen,"Port: %d; AdsState: %d; DeviceState: %d\n",Addr.port, nAdsState, nDeviceState);
      }
    }

    nErr = AdsPortCloseEx(nPort);
    if ( nErr )
    {
      slen = strlen(result);
      strncat(result,"AdsPortCloseEx() failed!",resultBufferSize-slen);
      return 0;
    }
    return 0;
  }

  return 0;
}

__declspec(dllexport) int filterTwinCAT(int *parameter, int accessFlag)
{
  return 0;
}
__declspec(dllexport) int initTwinCATcdi(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  cdi_mutex_t tmpMutex;
  int lastError, error = 0;
  
  /* try to get Mutex, is this a first application  */
  lastError = cdiMutexInit(&tmpMutex, cdiLine);
  if ( tmpMutex == NULL ) return -ask_error;
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);

  if (busLine<1 || busLine>nTwinCATLines) error = -illegal_line; 
//  error = initTwinCAT(busLine);

  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
  if ( error == 0 ) twincatInfo[busLine-1].cdiLine = cdiLine;
  return error;
}

// if sync call, data not pass filter then repeat Transmit
__declspec(dllexport) int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 10ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}

__declspec(dllexport) void execTwinCAT(CdiRequestInfoBlk *pReq)
{
  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }
  /* do first part; */
  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  do
  {
    doTwinCATRequest(pReq, 1);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest )
  {
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }

  /* do second part */
  do
  {
    doTwinCATRequest(pReq, 2);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }
    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag ) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  pReq->active = finishRequest; /* finish request */
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
}

// ln: 0 to  nTwinCATLines - 1
__declspec(dllexport) int initTwinCAT(int ln, CdiRequestInfoBlk *pReq)
{ int i, portNumber;

  printf("Twincat input devParameters: %d;%d; == %d; %d; %d, %d, %d, %d, %d\n",pReq->pDev->cdiAddr, pReq->pDev->cdiSubAddr, pReq->devParameters[0],pReq->devParameters[1],pReq->devParameters[2],pReq->devParameters[3],pReq->devParameters[4],pReq->devParameters[5],pReq->devParameters[6]);

  // Open communication port for local PLC (Run-time system 1)
  twincatInfo[ln].nPort = AdsPortOpenEx();
  if ( twincatInfo[ln].nPort == 0 )
  {
    printf("initTwinCAT error: AdsPortOpenEx() failed!; bus Line: %d\n", ln+1);
    return -not_ready;
  }
  // check netId input
  for (i=0; i<6; i++) if ( pReq->devParameters[paraNetId + i] ) break;
  if ( i != 6 ) // use input ip address
    for (i=0; i<6; i++) twincatInfo[ln].Addr.netId.b[i] = (char)pReq->devParameters[paraNetId + i];
  else
  {
    twincatInfo[ln].nErr = AdsGetLocalAddressEx(twincatInfo[ln].nPort, &(twincatInfo[ln].Addr));
    if ( twincatInfo[ln].nErr )
    {
      printf("initTwinCAT error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
      AdsPortCloseEx(twincatInfo[ln].nPort);
      return -not_ready;
    }
  }
  if ( pReq->devParameters[paraPortNumber] < 1 ) pReq->devParameters[paraPortNumber] = 1; // use first sps port
  if ( pReq->devParameters[paraPortNumber] >= 1 && pReq->devParameters[paraPortNumber] <= 4 ) // use PLC port number
  {
    if ( pReq->devParameters[paraPortNumber] == 1 ) portNumber = AMSPORT_R0_PLC_RTS1;
    else if ( pReq->devParameters[paraPortNumber] == 2 ) portNumber = AMSPORT_R0_PLC_RTS2;
    else if ( pReq->devParameters[paraPortNumber] == 3 ) portNumber = AMSPORT_R0_PLC_RTS3;
    else portNumber = AMSPORT_R0_PLC_RTS4;
  }
  else portNumber = pReq->devParameters[paraPortNumber];

  twincatInfo[ln].Addr.port = portNumber;

  twincatInfo[ln].pAddr = &(twincatInfo[ln].Addr);

  twincatInfo[ln].initFlag = 1;

  printf("Twincat portParameters: %d;%d; == %d; %d; %d, %d, %d, %d\n",twincatInfo[ln].nPort, portNumber, twincatInfo[ln].Addr.netId.b[0], twincatInfo[ln].Addr.netId.b[1],twincatInfo[ln].Addr.netId.b[2],twincatInfo[ln].Addr.netId.b[3],twincatInfo[ln].Addr.netId.b[4],twincatInfo[ln].Addr.netId.b[5]);

  return 0;
}

__declspec(dllexport) int exitTwinCAT(int ln)
{
  if ( twincatInfo[ln].initFlag == 0 ) return 0;
  twincatInfo[ln].nErr = AdsPortCloseEx(twincatInfo[ln].nPort);
  if ( twincatInfo[ln].nErr )  printf("initTwinCAT error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
  return 0;
}

int waitTwinCATfinished()
{ int ln, i;
  ULONG lHdlVar; 
  CdiLineInfoBlk *busHandle;
  for (ln=0; ln<N_TWINCAT_LINES; ln++)
  {
    if ((busHandle=cdiGetBusHandle(twincatInfo[ln].cdiLine)) == NULL) return -1;  /* something wrong!!  */
    cdiMutexLock(&(busHandle->lineMutex));

    for (i=0; i<busHandle->numberDevices; i++)
      if ( busHandle->pDevList[i].cdiSubAddr < 0 || busHandle->pDevList[i].cdiSubAddr > 1 )
      {
        lHdlVar = busHandle->pDevList[i].cdiSubAddr;
        twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_RELEASEHND, 0, sizeof(lHdlVar), &lHdlVar); 
      }
    exitTwinCAT(ln);
    cdiMutexUnlock(&(busHandle->lineMutex));
  }
  return 0;
}

int modifyAdsName(CdiRequestInfoBlk *pReq)
{
  char *pdest, *pdest2;
  char result[64];
  int ch, number, ret;

  // parse target astring: xxxx[]xxxxxx or xxxx[23]xxxxx
  ch ='[';
  pdest = strchr( pReq->pDev->lngName, ch );
  if ( pdest == NULL ) return 0;  // nothing to do

  ch =']';
  pdest2 = strchr( pReq->pDev->lngName, ch );
  if ( pdest2 == NULL ) return 0;  // nothing to do

  if ( pdest2 - pdest > 1 )
  {
    ret = sscanf(pdest,"[%d",&number);
    if ( ret == 1 ) return 0;   // nothing to do
  }
  strncpy(result,pReq->pDev->lngName, pdest - pReq->pDev->lngName);
  sprintf(&result[pdest - pReq->pDev->lngName],"[%d%s", pReq->pDev->cdiAddr, pdest2);
  strcpy(pReq->pDev->lngName, result);
  return 1;
}

__declspec(dllexport) int doTwinCATRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ln, ret, locTime, secTime;
  ULONG lHdlVar, numBytesRead; 

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? CA_WRITE : CA_READ;
  else /* second */
    access = pReq->accessFlag == cdiReadWriteFlag ? CA_WRITE : CA_READ;

  //printf("try to get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);
//  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  //printf("get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);

  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;

  if ( twincatInfo[ln].initFlag == 0 ) initTwinCAT(ln, pReq);

  // Fetch handle for an <szVar> PLC variable
  if ( pReq->pDev->cdiSubAddr == 0 )
  {
    ret = modifyAdsName(pReq);
    pReq->pDev->cdiSubAddr = 1;
  }
  locTime = getTimeMilliseconds();
  
// twincatInfo[ln].nErr = AdsSyncReadWriteReqEx2(twincatInfo[ln].nPort, twincatInfo[ln].pAddr, ADSIGRP_SYM_HNDBYNAME, 0x0, sizeof(lHdlVar), &lHdlVar, sizeof(pReq->pDev->lngName), pReq->pDev->lngName, &numBytesRead);
  if ( pReq->pDev->cdiSubAddr == 1 )  // user should set cdiSubAddr: 0 or 1 (only these two values !!) 2007-8-17
  {
    twincatInfo[ln].nErr = AdsSyncReadWriteReqEx2(twincatInfo[ln].nPort, twincatInfo[ln].pAddr, ADSIGRP_SYM_HNDBYNAME, 0x0, sizeof(lHdlVar), &lHdlVar, sizeof(pReq->pDev->lngName), pReq->pDev->lngName, &numBytesRead);
      printf("get symbole for: %s; value: %d; error number: %d\n",pReq->pDev->lngName, lHdlVar, twincatInfo[ln].nErr);
    if ( twincatInfo[ln].nErr == 0 )
    {
      pReq->pDev->cdiSubAddr = lHdlVar; // store handle
    }
  }
  else
  {
    twincatInfo[ln].nErr = 0;
    lHdlVar = pReq->pDev->cdiSubAddr;
  }

  secTime = getTimeMilliseconds();
  if ( secTime - locTime > reqHandleTime )
  {
    reqHandleTime = secTime - locTime;
    cdilog("ADS reqHandleTime(ms): %d\n",reqHandleTime);
  }
  if ( twincatInfo[ln].nErr )
  {
    printf("AdsSyncReadWriteReq error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
    pReq->errorCode = sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->pDev->numberDeviceError++;
    pReq->active = finishRequest; /* finish request */
  }
  else
  {
    if ( access == CA_READ )      // Read value of a PLC variable (by handle)
    {
      if ( pReq->pDev->dataFormat >= 0 && (unsigned)getTypeLength((short)pReq->pDev->dataFormat) < pReq->inLength )
      {
        twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->inLength, (char *)pReq->pInData, &numBytesRead );
/*
        dataLen = getTypeLength((short)pReq->pDev->dataFormat);
        numRead = pReq->inLength / dataLen;
        for ( i=0; i<numRead; i++ )
          twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, dataLen, ((char *)pReq->pInData) + dataLen * i, &numBytesRead );
*/
      } else
        twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->inLength, pReq->pInData, &numBytesRead );
    }
    else
    {
      if ( pReq->pDev->dataFormat >= 0 && (unsigned)getTypeLength((short)pReq->pDev->dataFormat) < pReq->outLength )
      {
        twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->outLength, (char *)pReq->pOutData);
/*
        dataLen = getTypeLength((short)pReq->pDev->dataFormat);
        numRead = pReq->outLength / dataLen;
        for ( i=0; i<numRead; i++ )
          twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, dataLen, ((char *)pReq->pOutData) + dataLen * i);
*/
      } else
        twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->outLength, pReq->pOutData);
    }
    if ( twincatInfo[ln].nErr )
    {
      pReq->errorCode = sedac_error;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      pReq->pDev->numberDeviceError++;
      pReq->active = finishRequest; /* finish request */
    }

    secTime = getTimeMilliseconds();
    if ( secTime - locTime > rwAdsTime )
    {
      rwAdsTime = secTime - locTime;
      cdilog("ADS reqHandleTime + rwTime (ms): %d\n",rwAdsTime);
    }

    //Handle der SPS-Variablen wieder freigeben
//2007-8-17
//    twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_RELEASEHND, 0, sizeof(lHdlVar), &lHdlVar); 

    secTime = getTimeMilliseconds();
    if ( secTime - locTime > totalAdsTime )
    {
      totalAdsTime = secTime - locTime;
      cdilog("ADS reqHandleTime + rwTime + closeHandle (ms): %d\n",totalAdsTime);
    }
  }

//  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return pReq->errorCode;
}

