#define CDI_RS232		// allocate memory

#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

#include "tinecdi.h"
#include "database.h"
#include "cdi.h"	// interface to user program
#include "cdiWinRS232Lib.h"

SER_STRUCT serInfo[N_SERIAL_LINES] = {
	{NULL,1}, {NULL,2}, {NULL,3}, {NULL,4} 
};

BAUD_STRUCT baudRate[] = {
	{110, CBR_110}, {300, CBR_300}, {600, CBR_600}, {1200, CBR_1200}, {2400, CBR_2400}, {4800, CBR_4800}, {9600, CBR_9600},
  {14400, CBR_14400}, {19200, CBR_19200}, {38400, CBR_38400}, {56000, CBR_56000}, {57600, CBR_57600}, 
  {115200, CBR_115200}, {128000, CBR_128000}, {256000, CBR_256000}
};
int numBaudRate = sizeof(baudRate) / sizeof(BAUD_STRUCT);


__declspec(dllexport) int filterRS232Bus(int *parameter, int accessFlag)
{
  return 0;
}

void rs232String(int cdiLine, char *str, unsigned int leng)
{ 
  CdiLineInfoBlk *busHandle;
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiRequestInfoBlk *pReq;
  /*
  if (NGdebug) 
    dbglog("call Back: cdi line: %d, port number: %d, string: %s, length: %d",
            cdiLine, busHandle[cdiLine]->lineNumber, str, leng);
  */
  if ((busHandle=cdiGetBusHandle(cdiLine)) == NULL)
  {
    cdilog("rs232String: Wrong cdi_line: %d for rs232String() call\n",cdiLine);
    return;
  }

  for ( busHandle->readCounter=0; busHandle->readCounter < busHandle->numberActiveRequest; busHandle->readCounter++ )
  {
    if ( busHandle->pReqList[ busHandle->readCounter ].active != recvRequest &&
    busHandle->pReqList[ busHandle->readCounter ].active != recvRequest2nd ) continue;
    pReq = &busHandle->pReqList[ busHandle->readCounter ];

	if ( pReq->inType == cdiChar )				// string input	*/
	{
	  if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
	  {
		if ( pReq->requestCounter == 1 ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* first call of Async or sync */
		else if ( pReq->patternFilter == NULL ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* no filter => any data */
		else if ( strcmp(str,pReq->pInData) != 0 ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* data changed compare to last one */
	  }
	  else /* only wait for specific pattern!! */
	  {
		/*if ( strcmp(str, pReq->patternFilter) == 0 ) */
		if ( strstr(str, pReq->patternFilter) != NULL )
		{
		  if ( pReq->conditionFlag == maskCompEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
		  /* else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; */
		}
	    else if ( pReq->conditionFlag == maskCompNotEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
		/* else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; */
	  }
    }
	else
	{
	  checkReturnCondition(pReq, pReq->pInData, str, leng, 1);
	}
    if ( pReq->dataReturnCondition == 0 ) continue;
    /*
    if (NGdebug) dbglog("rs232 return condition OK: %s",str);
    */
    if ( leng > pReq->inLength ) strncpy( pReq->pInData, str, pReq->inLength);
    else strcpy(pReq->pInData, str);

    cdiExecRequest(&(busHandle->pReqList[busHandle->readCounter]), readTaskIndex);
  }
}

/****************************************/
/*	send String							*/
/****************************************/
unsigned int sendSerialPort(SER_STRUCT *pSer, char *psOutput)
{
  if ( pSer->serFd == NULL )
  {
    cdilog(" port is not open yet! port: com%d\n", pSer->channel+1);
    return -file_not_open;
  }

  pSer->wrtRet = ResetEvent(pSer->overlap[1].hEvent);
  pSer->wrtRet = WriteFile(pSer->serFd, // handle to file
    (LPCVOID)psOutput,          // data buffer
    (DWORD)strlen(psOutput),    // number of bytes to write
    (LPDWORD)&pSer->iOut,             // number of bytes written
    (LPOVERLAPPED)&pSer->overlap[1]          // overlapped buffer
  );
  if ( pSer->wrtRet == 0 && GetLastError() == ERROR_IO_PENDING )
  {
   	pSer->wrtRet = WaitForSingleObject (pSer->overlap[1].hEvent, INFINITE);
    if ( pSer->wrtRet != WAIT_OBJECT_0 )
    {
      cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, pSer->wrtRet);
      return -rs232_error;
    }
    pSer->wrtRet = GetOverlappedResult(
  			pSer->serFd,				// Handle to device
	  		&pSer->overlap[1],	// pointer to overlapped structure  
		  	&pSer->iOut,// pointer to actual bytes count  
  			FALSE );// wait flag!
    if ( pSer->wrtRet == 0 )
    {
      cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, GetLastError());
      return -rs232_error;
    }
  }
  else
  {
    pSer->wrtRet = GetLastError();
    cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, pSer->wrtRet);
    return -rs232_error;
  }

  if (pSer->iOut < strlen(psOutput))
  {
    cdilog("write data to serial port error: written: %d; requested: %d\n", pSer->iOut, strlen(psOutput));
  }
  else printf("write to com%d ==> wrote %d chars: %s\n", pSer->channel, pSer->iOut, psOutput);

  return pSer->iOut;
}

/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
__declspec(dllexport) void execRS232(CdiRequestInfoBlk *pReq)
{
  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }
  if ( pReq->active == transRequest ) /* start first transfer */
  {
    if ( pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag ) /* write first */
    {
      if ( sendSerialPort(&serInfo[ ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1 ], pReq->pOutData) != pReq->outLength )
      {
        pReq->errorCode = rs232_error; /* write failed */
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
        pReq->active = finishRequest;
      } /* finish or 2ndReceive */
      else pReq->active = pReq->accessFlag == cdiWriteFlag ? finishRequest : recvRequest2nd;
    }
    else pReq->active = recvRequest; /* jump to receive */
    return;
  }
  /* second transfer */
  if ( pReq->accessFlag == cdiReadWriteFlag ) /* 2nd write */
  {
    if ( sendSerialPort(&serInfo[ ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1], pReq->pOutData) != pReq->outLength )
    {
      pReq->errorCode = rs232_error;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    }
    pReq->active = finishRequest;
  }

  return;
}

/********************************************/
/*	receive string from port				*/
/********************************************/
//int receiveSerialPortString(int channelNumber, char *pString)
DWORD WINAPI GetSerialStringThread(SER_STRUCT *pSer)
{
  do
  {
 	  pSer->ret = ResetEvent(pSer->overlap[0].hEvent);
    pSer->ret = ReadFile(pSer->serFd,   // handle to file
                    (LPVOID)pSer->buf,  // data buffer
                    STRING_LENGTH,             // number of bytes to read
                    (LPDWORD)&pSer->length,    // number of bytes read
                    &pSer->overlap[0]           // overlapped buffer
                  );
	  if ( pSer->ret == 0 && GetLastError() == ERROR_IO_PENDING )
    {
    	pSer->ret = WaitForSingleObject(pSer->overlap[0].hEvent, INFINITE);
      if ( pSer->ret != WAIT_OBJECT_0 )
      {
        cdilog("ReadFile(WaitForSingleObject) failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
        continue;
      }
      pSer->ret = GetOverlappedResult(
  			pSer->serFd,				// Handle to device
	  		&pSer->overlap[0],	// pointer to overlapped structure  
		  	&pSer->length,      // pointer to actual bytes count  
  			FALSE );            // wait flag!
      if ( pSer->ret == 0 )
      {
        cdilog("ReadFile(GetOverlappedResult) failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
        continue;
      }
    }
    else
    {
      cdilog("ReadFile Call failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
      continue;
    }

    if ( pSer->length <= 0 )
    {
      cdilog("ReadFile read wrong data, length; on port com%d; read length: %d\n", pSer->channel+1, pSer->length);
      continue;
    }
    pSer->buf[pSer->length] = 0;
    printf("comport: com%d; input string: %s\n", pSer->channel+1, pSer->buf);
    cdilog("comport: com%d; input string: %s\n", pSer->channel+1, pSer->buf);
	  if (pSer->receiveString) pSer->receiveString(pSer->cdiLine, pSer->buf, pSer->length);
  } while (1);
}

/****************************************/
/* open port 0-3(com1 to com4)			*/
/* parameters[], see header file		*/
/****************************************/
int openSerialPort(int channelNumber, int *parameters, void (*callBack)(int, char*, unsigned int), int cdiLine)
{ char pcCommPort[16];
  HANDLE hCom;
  DCB dcb;
  COMMTIMEOUTS commtmouts;
  BOOL fSuccess;
  SER_STRUCT *pSer;
  DWORD dwCdiThreadId;
  int i;

  if ( channelNumber < 1 || channelNumber > N_SERIAL_LINES ) return -illegal_line;
  channelNumber--;

  // find correct baudrate
  for (i=0; i<numBaudRate; i++) if (parameters[paramBaudrate] == baudRate[i].inputBaud) break;
  if ( i == numBaudRate )
  {
    cdilog("input wrong! baudrate: %d!\n", parameters[paramBaudrate]);
	  return -invalid_data;
  }

  pSer = &serInfo[channelNumber];
  if ( pSer->serFd != NULL )
  {
    cdilog("serial already opened! line: %d! for Com%d\n", channelNumber, channelNumber+1);
    return 0;
  }

  // create handle
  sprintf( pcCommPort, "COM%d", channelNumber+1 );
  hCom = CreateFile( pcCommPort,
                     GENERIC_READ | GENERIC_WRITE,
                     0,    // comm devices must be opened w/exclusive-access
                     NULL, // no security attributes
                     OPEN_EXISTING, // comm devices must use OPEN_EXISTING
                     FILE_FLAG_OVERLAPPED,    // not overlapped I/O
                     NULL  // hTemplate must be NULL for comm devices
                     );

  if (hCom == INVALID_HANDLE_VALUE) {      // Handle the error.
    cdilog("CreateFile for %s failed with error 0x%x.\n", pcCommPort, GetLastError());
    return -file_not_open;
  }

  // set buffer size
  fSuccess = SetupComm(
    hCom,           // handle to communications device
    dwInQueue,      // size of input buffer
    dwOutQueue      // size of output buffer
  );
  if (!fSuccess) {      // Handle the error.
    cdilog("SetupComm for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }

  // We will build on the current configuration, and skip setting the size
  // of the input and output buffers with SetupComm.
  fSuccess = GetCommState(hCom, &dcb);
  if (!fSuccess) {      // Handle the error.
    cdilog("GetCommState for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }
  
  memcpy(&pSer->dcbCopy, &dcb, sizeof(DCB));

  // Fill in the DCB: baud=57,600 bps, 8 data bits, no parity, and 1 stop bit.
  dcb.BaudRate = parameters[paramBaudrate];;     // set the baud rate
  dcb.ByteSize = parameters[param8bits] == 0 ? 7 : 8;           // data size, xmit, and rcv
  dcb.fParity = parameters[paramParity] ? TRUE : FALSE;         // no parity bit
  dcb.StopBits = parameters[param2StopBits] ? TWOSTOPBITS : ONESTOPBIT;    // one stop bit
  if ( dcb.fParity ) dcb.Parity = parameters[paramOddParity] ? ODDPARITY : EVENPARITY;
  else dcb.Parity = NOPARITY;
  if ( parameters[paramFlowControl] )
  {
    dcb.fRtsControl = RTS_CONTROL_ENABLE;
    dcb.fOutxCtsFlow = TRUE;
  }

  fSuccess = SetCommState(hCom, &dcb);
  if (!fSuccess) {     // Handle the error.
    cdilog("SetCommState for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }

  if ( (fSuccess = GetCommTimeouts(hCom, &commtmouts)) == 0 )
  {      // Handle the error.
    cdilog("GetCommTimeouts for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    CloseHandle(hCom);
    return -rs232_error;
  }
  memcpy(&pSer->commtmoutsCopy, &commtmouts, sizeof(COMMTIMEOUTS));
  commtmouts.ReadIntervalTimeout = 1000 * 10 * 20 / parameters[paramBaudrate];  // timeout in ms for 10 char transmission, 10 bit per char
  commtmouts.ReadTotalTimeoutConstant = 0;  // timeout not used!
  commtmouts.ReadTotalTimeoutMultiplier = 0;
//  commtmouts.WriteTotalTimeoutConstant = 1;
//  commtmouts.WriteTotalTimeoutMultiplier = 1;
  if ( (fSuccess = SetCommTimeouts(hCom, &commtmouts)) == 0 )
  {     // Handle the error.
    cdilog("SetCommTimeouts for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    CloseHandle(hCom);
    return -rs232_error;
  }

  // create event for read and write
  for (i=0; i<2; i++)
  {
    pSer->overlap[i].Offset = 0;
		pSer->overlap[i].OffsetHigh = 0;
    pSer->overlap[i].hEvent = CreateEvent(
						NULL,	// pointer to security attributes
						TRUE,	// flag for manual-reset event
						FALSE,	// flag for initial state(false: nonsignaled)
						NULL );	// pointer to event-object name
    if ( pSer->overlap[i].hEvent == NULL ) break;
  }
  if ( i < 2 )
  {
    if ( i == 1 ) CloseHandle( pSer->overlap[i-1].hEvent );
    cdilog("CreateEvent for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    fSuccess = SetCommTimeouts(hCom, &pSer->commtmoutsCopy);
    CloseHandle(hCom);
  	return -rs232_error;
	}

	// Waiting for messages and work them out on master module
  if ( (pSer->tIdMaster = CreateThread(NULL,0,GetSerialStringThread,(void *)pSer,0,&dwCdiThreadId)) == NULL )
  {
	  cdilog ("CreateThread line: %s error",pcCommPort);
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    fSuccess = SetCommTimeouts(hCom, &pSer->commtmoutsCopy);
    for ( i=0; i<2; i++ ) CloseHandle( pSer->overlap[i].hEvent );
    CloseHandle(hCom);
	  return -un_allocated;
  }

  pSer->baudRate = parameters[paramBaudrate];
  pSer->usOneChar = 1000000 / pSer->baudRate * 10;	// 10 bit per char
  pSer->channel = channelNumber;
  pSer->cdiLine = cdiLine;
  pSer->receiveString = callBack;
  pSer->canonicalMode = parameters[paramCanonicalInput];

  cdilog("Serial port %s successfully reconfigured. handle: %d\n", pcCommPort, hCom);

  pSer->serFd = hCom;
  fSuccess = PurgeComm(pSer->serFd, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
  if ( fSuccess == 0 ) cdilog("PurgeComm error! GetLastError(): 0x%x\n",GetLastError());

  return 0;
}

/****************************************/
#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}
__declspec(dllexport) int initRS232ForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return openSerialPort(busLine, parameter, rs232String, cdiLine);
}

/****************************************/
/*	Close								*/
/****************************************/
__declspec(dllexport) int closeSerialPort(int channelNumber)
{ SER_STRUCT *pSer;
  BOOL fSuccess;
  int i;

  if ( channelNumber < 1 || channelNumber > N_SERIAL_LINES ) return -illegal_line;

  cdilog("close serialPort! channelNumber: Com-%d\n", channelNumber);

  pSer = &serInfo[--channelNumber];
  if ( pSer->serFd == NULL ) return -file_not_open;
  CancelIo(pSer->serFd);
  CloseHandle( pSer->tIdMaster );
  fSuccess = SetCommState(pSer->serFd, &pSer->dcbCopy);
  fSuccess = GetCommTimeouts(pSer->serFd, &pSer->commtmoutsCopy);
  for(i=0;i<2;i++) CloseHandle( pSer->overlap[i].hEvent );
  CloseHandle(pSer->serFd);
  pSer->serFd = NULL;

  return 0;
}
